"""
Serial Communications
The implementation of RTU reception driver may imply the management of a lot of interruptions due to the t1.5 and t3.5
timers. With high communication baud rates, this leads to a heavy CPU load. Consequently these two timers must be
strictly respected when the baud rate is equal or lower than 19200 Bps. For baud rates greater than 19200 Bps,
fixed values for the 2 timers should be used: it is recommended to use a value of 750µs for the inter-character
time-out (t1.5) and a value of 1.750ms for inter-frame delay (t3.5).
Source www.modbus.org Modbus_over_serial_line_V1.02 2006
"""
import asyncio
import struct
from dataclasses import dataclass
import logging
from aiomodbus import decoders, encoders
import serial_asyncio

import aiomodbus.crc
from aiomodbus.exceptions import modbus_exception_codes

log = logging.getLogger(__file__)


class ModbusSerialProtocol(asyncio.Protocol):
    def __init__(self, client):
        self.transport = None
        self.client = client
        self.current_request = None
        self.recv_callback = None
        self.q = asyncio.Queue()
        self.buffer = bytearray()
        self.loop = asyncio.get_event_loop()
        self.byte_time = 0

    def connection_made(self, transport: serial_asyncio.SerialTransport):
        self.transport = transport
        self.byte_time = 1 / (
            transport.serial.baudrate
            / (transport.serial.bytesize + transport.serial.stopbits + 1)
            / 1.5
        )
        self.client.connected.set()

    def data_received(self, data):
        self.buffer.extend(data)
        self.q.put_nowait(None)

    async def decode(
        self,
        packet_length: int,
        decode_packing: str,
        turn_around_delay_timeout: float = 0.4,
        timeout: float = 0.1,
    ) -> tuple:
        try:
            await asyncio.wait_for(self.q.get(), turn_around_delay_timeout)
            return await asyncio.wait_for(
                self.build_decode(packet_length, decode_packing),
                self.byte_time * packet_length + timeout,
            )
        finally:
            await self.empty_queue()

    async def empty_queue(self):
        await asyncio.sleep(self.byte_time * 2.5)
        while True:
            try:
                self.q.get_nowait()
            except asyncio.QueueEmpty:
                break

    async def build_decode(self, packet_length: int, decode_packing: str) -> tuple:
        while True:
            if len(self.buffer) >= 5:
                if self.buffer[1] & 0x80:
                    log.debug(
                        "Decode: " + " ".join(f"{byt:02X}" for byt in self.buffer)
                    )
                    raise modbus_exception_codes[self.buffer[2]]
            if len(self.buffer) >= packet_length:
                aiomodbus.crc.check_crc(self.buffer[:packet_length])
                log.debug(
                    "Decode: "
                    + " ".join(f"{byt:02X}" for byt in self.buffer[:packet_length])
                )
                return struct.unpack(decode_packing, self.buffer[:packet_length])
            await self.q.get()

    def connection_lost(self, exc):
        self.client.connected.clear()


@dataclass
class ModbusSerialClient:
    port: str
    baudrate: int
    parity: str
    stopbits: int
    bytesize: int = 8
    default_unit_id: int = 0
    transport: asyncio.Transport = None
    protocol: ModbusSerialProtocol = None

    def __post_init__(self):
        self.transaction = asyncio.Lock()
        self.t_1_5 = None
        self.t_3_5 = None
        self.connected = asyncio.Event()

    async def connect(self):
        self.transport, self.protocol = await serial_asyncio.create_serial_connection(
            asyncio.get_running_loop(),
            lambda: ModbusSerialProtocol(self),
            url=self.port,
            baudrate=self.baudrate,
            parity=self.parity,
            stopbits=self.stopbits,
            bytesize=self.bytesize,
        )
        await self.connected.wait()

    def _encode_packet(self, unit, function_code, address, *values) -> bytearray:
        packet = bytearray()
        packet.extend(struct.pack(">BB", unit, function_code))
        packet.extend(encoders.from_func_code(function_code, address, *values))
        packet.extend(struct.pack(">H", aiomodbus.crc.calc_crc(packet)))
        log.debug(("Encode: " + " ".join(f"{byt:02X}" for byt in packet)))
        return packet

    def _pack_bits(self, *values: bool, size=8):
        vals = [0] * (len(values) // size + 1)
        for ind, bit in enumerate(values):
            vals[ind // size] += bit << ind % size
        return vals

    def _upack_bits(self, *values: int, size=8):
        vals = []
        for val in values:
            for ind in range(size):
                vals.append(bool((val >> ind) & 1))
        return vals

    async def _request(
        self,
        unit: int,
        function_code: int,
        address: int,
        *values: int,
        decode_packing: str,
        packet_length: int,
        turn_around_delay_timeout: float = 0.4,
    ):
        if unit is None:
            unit = self.default_unit_id
        async with self.transaction:
            try:
                await asyncio.wait_for(self.connected.wait(), 2)
            except asyncio.TimeoutError as e:
                raise asyncio.TimeoutError(
                    "Failed modbus request as client is not connected"
                ) from e
            packet = self._encode_packet(unit, function_code, address, *values)
            self.protocol.buffer.clear()
            self.transport.write(packet)
            write_time = self.protocol.byte_time * len(packet)
            unit_id, func_code, *values, crc = await self.protocol.decode(
                packet_length,
                decode_packing,
                turn_around_delay_timeout=turn_around_delay_timeout + write_time,
            )
            assert unit_id == unit
            assert function_code == func_code
            return values

    async def read_coils(
        self, address: int, count: int, *, unit=None, timeout: float = 0.1
    ):
        resp = await self._request(
            unit,
            0x01,
            address,
            count,
            decode_packing=">BBB" + "B" * ((count - 1) // 8 + 1) + "H",
            packet_length=5 + 1 * ((count - 1) // 8 + 1),
            turn_around_delay_timeout=timeout,
        )
        return self._upack_bits(*resp[1:])[:count]

    async def read_discrete_inputs(
        self, address: int, count: int, *, unit=None, timeout: float = 0.1
    ):
        resp = await self._request(
            unit,
            0x02,
            address,
            count,
            decode_packing=">BBB" + "B" * ((count - 1) // 8 + 1) + "H",
            packet_length=5 + 1 * ((count - 1) // 8 + 1),
            turn_around_delay_timeout=timeout,
        )
        return self._upack_bits(*resp[1:])[:count]

    async def read_holding_registers(
        self, address: int, count: int, *, unit=None, timeout: float = 0.1
    ):
        resp = await self._request(
            unit,
            0x03,
            address,
            count,
            decode_packing=">BBBH" + "H" * count,
            packet_length=5 + 2 * count,
            turn_around_delay_timeout=timeout,
        )
        return resp[1:]

    async def read_input_registers(
        self, address: int, count: int, *, unit=None, timeout: float = 0.1
    ):
        resp = await self._request(
            unit,
            0x04,
            address,
            count,
            decode_packing=">BBBH" + "H" * count,
            packet_length=5 + 2 * count,
            turn_around_delay_timeout=timeout,
        )
        return resp[1:]

    async def write_single_coil(
        self, address: int, value: bool, *, unit=None, timeout: float = 0.1
    ):
        if value:
            value = 0xFF00
        await self._request(
            unit,
            0x05,
            address,
            value,
            decode_packing=">BBHHH",
            packet_length=8,
            turn_around_delay_timeout=timeout,
        )

    async def write_single_register(
        self, address: int, value: int, *, unit=None, timeout: float = 0.1
    ):
        return await self._request(
            unit,
            0x06,
            address,
            value,
            decode_packing=">BBHHH",
            packet_length=8,
            turn_around_delay_timeout=timeout,
        )

    async def write_multiple_coils(
        self, address: int, *values: bool, unit=None, timeout: float = 0.1
    ):
        await self._request(
            unit,
            0x0F,
            address,
            *values,
            decode_packing=">BBHHH",
            packet_length=8,
            turn_around_delay_timeout=timeout,
        )

    async def write_multiple_registers(
        self, address: int, *values: int, unit=None, timeout: float = 0.1
    ):
        await self._request(
            unit,
            0x10,
            address,
            *values,
            decode_packing=">BBHHH",
            packet_length=8,
            turn_around_delay_timeout=timeout,
        )

    async def read_exception_status(self, unit=None, timeout=None):
        function_code = 0x07
        raise NotImplementedError

    async def diagnostics(self, sub_function, *data, unit=None, timeout=None):
        function_code = 0x08
        raise NotImplementedError



import pytest
import struct
from unittest.mock import MagicMock
import aiomodbus
import asyncio

import aiomodbus.exceptions
import aiomodbus.serial
import aiomodbus.tcp


def async_return(result):
    f = asyncio.Future()
    f.set_result(result)
    return f


def respond(protocol, arr):
    def _tmp(data):
        protocol.data_received(arr)

    return _tmp


@pytest.fixture
def create_mock_coro(mocker, monkeypatch):
    def _create_mock_patch_coro(to_patch=None):
        mock = mocker.Mock()

        async def _coro(*args, **kwargs):
            return mock(*args, **kwargs)

        if to_patch:  # <-- may not need/want to patch anything
            monkeypatch.setattr(to_patch, _coro)
        return mock, _coro

    return _create_mock_patch_coro


@pytest.fixture
def mock_sleep(create_mock_coro):
    # won't need the returned coroutine here
    mock, _ = create_mock_coro(to_patch="asyncio.sleep")
    return mock


@pytest.fixture
def client():
    client = aiomodbus.serial.ModbusSerialClient("COM3", 9600, "N", 1)
    client.transport = MagicMock()
    client.protocol = aiomodbus.serial.ModbusSerialProtocol(client)
    client.protocol.byte_time = 0.1
    client.connected.set()
    return client


@pytest.mark.asyncio
async def test_read_coils(mock_sleep, client):
    client.transport.write.side_effect = respond(
        client.protocol, b"\x11\x01\x05\xCD\x6B\xB2\x0E\x1B\x45\xE6"
    )
    fut = asyncio.create_task(client.read_coils(0x13, 0x25, unit=0x11))
    await fut
    client.transport.write.assert_called_once_with(b"\x11\x01\x00\x13\x00\x25\x0E\x84")
    assert fut.result() == [
        True,
        False,
        True,
        True,
        False,
        False,
        True,
        True,
        True,
        True,
        False,
        True,
        False,
        True,
        True,
        False,
        False,
        True,
        False,
        False,
        True,
        True,
        False,
        True,
        False,
        True,
        True,
        True,
        False,
        False,
        False,
        False,
        True,
        True,
        False,
        True,
        True,
    ]


@pytest.mark.asyncio
async def test_read_discrete_inputs(mock_sleep, client):
    client.transport.write.side_effect = respond(
        client.protocol, b"\x11\x02\x03\xAC\xDB\x35\x20\x18"
    )
    fut = asyncio.create_task(client.read_discrete_inputs(0xC4, 0x16, unit=0x11))
    await fut
    client.transport.write.assert_called_once_with(b"\x11\x02\x00\xC4\x00\x16\xBA\xA9")
    assert fut.result() == [
        False,
        False,
        True,
        True,
        False,
        True,
        False,
        True,
        True,
        True,
        False,
        True,
        True,
        False,
        True,
        True,
        True,
        False,
        True,
        False,
        True,
        True,
    ]


@pytest.mark.asyncio
async def test_read_holding_registers(mock_sleep, client):
    client.transport.write.side_effect = respond(
        client.protocol, b"\x11\x03\x06\xAE\x41\x56\x52\x43\x40\x49\xAD"
    )
    fut = asyncio.create_task(client.read_holding_registers(0x6B, 0x3, unit=0x11))
    await fut
    client.transport.write.assert_called_once_with(b"\x11\x03\x00\x6b\x00\x03\x76\x87")
    assert fut.result() == [0xAE41, 0x5652, 0x4340]


@pytest.mark.asyncio
async def test_read_input_registers(mock_sleep, client):
    client.transport.write.side_effect = respond(
        client.protocol, b"\x11\x04\x02\x00\x0A\xF8\xF4"
    )
    fut = asyncio.create_task(client.read_input_registers(0x08, 0x1, unit=0x11))
    await fut
    client.transport.write.assert_called_once_with(b"\x11\x04\x00\x08\x00\x01\xB2\x98")
    assert fut.result() == [0xA]


@pytest.mark.asyncio
async def test_write_coil(mock_sleep, client):
    client.transport.write.side_effect = respond(
        client.protocol, b"\x11\x05\x00\xAC\xFF\x00\x4E\x8B"
    )
    fut = asyncio.create_task(client.write_single_coil(0xAC, True, unit=0x11))
    await fut
    client.transport.write.assert_called_once_with(b"\x11\x05\x00\xAC\xFF\x00\x4E\x8B")


@pytest.mark.asyncio
async def test_write_register(mock_sleep, client):
    client.transport.write.side_effect = respond(
        client.protocol, b"\x11\x06\x00\x01\x00\x03\x9A\x9B"
    )
    fut = asyncio.create_task(client.write_single_register(0x01, 0x3, unit=0x11))
    await fut
    client.transport.write.assert_called_once_with(b"\x11\x06\x00\x01\x00\x03\x9A\x9B")


@pytest.mark.asyncio
async def test_write_coils(mock_sleep, client):
    client.transport.write.side_effect = respond(
        client.protocol, b"\x11\x0F\x00\x13\x00\x0A\x26\x99"
    )
    fut = asyncio.create_task(
        client.write_multiple_coils(
            0x13,
            True,
            False,
            True,
            True,
            False,
            False,
            True,
            True,
            True,
            False,
            unit=0x11,
        )
    )
    await fut
    client.transport.write.assert_called_once_with(
        b"\x11\x0F\x00\x13\x00\x0A\x02\xCD\x01\xBF\x0B"
    )


@pytest.mark.asyncio
async def test_write_registers(mock_sleep, client):
    client.transport.write.side_effect = respond(
        client.protocol, b"\x11\x10\x00\x01\x00\x02\x12\x98"
    )
    fut = asyncio.create_task(
        client.write_multiple_registers(0x01, 0xA, 0x102, unit=0x11)
    )
    await fut
    client.transport.write.assert_called_once_with(
        b"\x11\x10\x00\x01\x00\x02\x04\x00\x0A\x01\x02\xC6\xF0"
    )


@pytest.mark.parametrize(
    "exceptioncls,exception_code",
    [
        (aiomodbus.exceptions.IllegalFunction, 1),
        (aiomodbus.exceptions.IllegalDataAddress, 2),
        (aiomodbus.exceptions.IllegalDataValue, 3),
        (aiomodbus.exceptions.SlaveDeviceFailure, 4),
        (aiomodbus.exceptions.AcknowledgeError, 5),
        (aiomodbus.exceptions.DeviceBusy, 6),
        (aiomodbus.exceptions.NegativeAcknowledgeError, 7),
        (aiomodbus.exceptions.MemoryParityError, 8),
        (aiomodbus.exceptions.GatewayPathUnavailable, 10),
        (aiomodbus.exceptions.GatewayDeviceFailedToRespond, 11),
        (ConnectionError, 12),
    ],
)
@pytest.mark.asyncio
async def test_exceptions(exceptioncls, exception_code, mock_sleep, client):
    exc_packet = bytearray([0x11, 0x83, exception_code])
    exc_packet.extend(struct.pack(">H", aiomodbus.crc.calc_crc(exc_packet)))
    client.transport.write.side_effect = respond(client.protocol, exc_packet)
    with pytest.raises(exceptioncls):
        await client.read_holding_registers(0x6B, 0x3, unit=0x11)
    client.transport.write.assert_called_once_with(b"\x11\x03\x00\x6b\x00\x03\x76\x87")


@pytest.mark.asyncio
async def test_read_exception_status(client):
    with pytest.raises(NotImplementedError):
        await client.read_exception_status()


@pytest.mark.asyncio
async def test_diagnostics(client):
    with pytest.raises(NotImplementedError):
        await client.diagnostics(None)



import struct


def _pack_words(address: int, *values: int) -> bytes:
    return struct.pack(">H" + "H" * len(values), address, *values)


def _pack_single_coil(address: int, value: int) -> bytes:
    if value:
        value = 0xFF00
    else:
        value = 0
    return struct.pack(">HH", address, value)


def _pack_write_coils(address: int, *values: bool) -> bytes:
    size = 8
    vals = [0] * (len(values) // size + 1)
    for ind, bit in enumerate(values):
        vals[ind // size] += bit << ind % size
    return struct.pack(">HHB" + "B" * len(vals), address, len(values), len(vals), *vals)


def _pack_write_words(address: int, *values: int) -> bytes:
    return struct.pack(
        ">HHB" + "H" * len(values), address, len(values), len(values) * 2, *values
    )


function_codes = {
    1: _pack_words,
    2: _pack_words,
    3: _pack_words,
    4: _pack_words,
    5: _pack_single_coil,
    6: _pack_words,
    15: _pack_write_coils,
    16: _pack_write_words,
}


def from_func_code(func_code, address, *data):
    return function_codes[func_code](address, *data)



import struct
from aiomodbus.exceptions import modbus_exception_codes


def _unpack_bits(data, bytesize=8):
    (size,) = struct.unpack(">B", data[0:1])
    values = struct.unpack(">" + "B" * size, data[1:])
    vals = []
    for val in values:
        for ind in range(bytesize):
            vals.append(bool((val >> ind) & 1))
    return vals


def _unpack_words(data):
    (size,) = struct.unpack(">B", data[0:1])
    return struct.unpack(">" + "H" * (size // 2), data[1:])


def _unpack_single_register(data):
    return struct.unpack(">H", data[2:4])[0]


def _unpack_single_coil(data):
    return bool(_unpack_single_register(data))


function_codes = {
    1: _unpack_bits,
    2: _unpack_bits,
    3: _unpack_words,
    4: _unpack_words,
    5: _unpack_single_coil,
    6: _unpack_single_register,
    15: _unpack_single_register,
    16: _unpack_single_register,
}


def from_func_code(fut, func_code, data):
    if func_code & 0x80:
        fut.set_exception(modbus_exception_codes[data[0]])
    else:
        try:
            fut.set_result(function_codes[func_code](data))
        except BaseException as e:
            fut.set_exception(e)



import pytest
import aiomodbus.crc


@pytest.mark.parametrize(
    "arr,result",
    [
        (bytearray(b"\x01\x03\x14" + b"\x00" * 20), 0xA367),
        (bytearray(b"\x01\x03\x00\x01\x00\x0A"), 0x940D),
    ],
)
def test_calc_crc(arr, result):
    assert aiomodbus.crc.calc_crc(arr) == result


@pytest.mark.parametrize(
    "arr",
    [
        bytearray(b"\x01\x03\x14" + b"\x00" * 20 + b"\xA3\x67"),
        bytearray(b"\x01\x03\x00\x01\x00\x0A\x94\x0D"),
    ],
)
def test_check_crc_pass(arr):
    aiomodbus.crc.check_crc(arr)


@pytest.mark.parametrize(
    "arr",
    [
        bytearray(b"\x01\x03\x14" + b"\x00" * 20 + b"\xA3\x67"),
        bytearray(b"\x01\x03\x00\x01\x00\x0A\x94\x0D"),
    ],
)
def test_check_crc_fail(arr):
    arr[-1] = (arr[-1] + 1) % 255
    with pytest.raises(aiomodbus.crc.CrcValidationError):
        aiomodbus.crc.check_crc(arr)



class RequestException(ValueError):
    pass


class IllegalFunction(RequestException):
    pass


class IllegalDataAddress(RequestException):
    pass


class IllegalDataValue(RequestException):
    pass


class MemoryParityError(IOError):
    pass


class SlaveDeviceFailure(IOError):
    pass


class AcknowledgeError(IOError):
    pass


class DeviceBusy(IOError):
    pass


class NegativeAcknowledgeError(IOError):
    pass


class GatewayPathUnavailable(IOError):
    pass


class GatewayDeviceFailedToRespond(IOError):
    pass


modbus_exception_codes = {
    1: IllegalFunction,
    2: IllegalDataAddress,
    3: IllegalDataValue,
    4: SlaveDeviceFailure,
    5: AcknowledgeError,
    6: DeviceBusy,
    7: NegativeAcknowledgeError,
    8: MemoryParityError,
    10: GatewayPathUnavailable,
    11: GatewayDeviceFailedToRespond,
    12: ConnectionError,
}



import asyncio
import logging
import socket
from typing import Optional, Tuple, Union, Dict
import struct
from dataclasses import dataclass
from asyncio import transports, Lock
from aiomodbus import decoders, encoders

Number = Union[int, float]

log = logging.getLogger(__file__)


@dataclass
class TransactionLimit:
    limit: Optional[int] = None
    evt_connected: Optional[asyncio.Event] = None
    evt_connected_timeout: float = 2

    def __post_init__(self):
        self.semaphore = None
        if self.limit:
            self.semaphore = asyncio.Semaphore(self.limit)

    async def __aenter__(self):
        if self.evt_connected:
            try:
                await asyncio.wait_for(
                    self.evt_connected.wait(), self.evt_connected_timeout
                )
            except asyncio.TimeoutError:
                raise ConnectionError("Client isn't connected")
        if self.semaphore:
            await self.semaphore.__aenter__()

    def __await__(self):
        if self.semaphore:
            return self.semaphore.__await__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.semaphore:
            await self.semaphore.__aexit__(exc_type, exc_val, exc_tb)


class ModbusTcpProtocol(asyncio.Protocol):
    def __init__(self, client):
        self.client = client
        self.transactions: Dict[int, asyncio.Future] = {}
        self._cnt_lock = Lock()
        self.transaction_cnt = 0

    async def next_transaction(self):
        async with self._cnt_lock:
            self.transaction_cnt = (self.transaction_cnt % 0xFFFE) + 1
            return self.transaction_cnt

    def connection_made(self, transport: transports.BaseTransport) -> None:
        log.info(f"Modbus Client connected at {self.client.host}")

    def data_received(self, data: bytes) -> None:
        while True:
            header, payload = data[:8], data[8:]
            trans_id, protocol_id, length, unit_id, func_code = struct.unpack(
                ">HHHBB", header
            )
            log.debug("Decode: " + " ".join(f"{byt:02X}" for byt in data))
            fut = self.transactions.get(trans_id)
            if fut:
                try:
                    decoders.from_func_code(fut, func_code, payload[: length - 2])
                    if len(payload) <= length - 2:
                        break
                except Exception as e:
                    log.exception(e)
                    break
            data = payload[length - 2 :]

    def connection_lost(self, exc: Optional[Exception]) -> None:
        self.client.connected.clear()
        log.info(f"Modbus Client disconnected from {self.client.host}")
        if self.client.running:
            for fut in self.transactions.values():
                if not fut.done():
                    fut.cancel()
            asyncio.create_task(self.client.connect())

    async def new_transaction(self) -> Tuple[int, asyncio.Future]:
        trans_id = await self.next_transaction()
        fut = asyncio.Future()
        self.transactions[trans_id] = fut
        return trans_id, fut


@dataclass
class ModbusTCPClient:
    host: str
    port: int = 502
    client_port: int = 0
    max_active_requests: Optional[int] = None
    default_unit_id: int = 0
    default_timeout: Optional[Number] = 0.2
    auto_reconnect_after: Optional[Number] = None
    transport: asyncio.Transport = None
    protocol: ModbusTcpProtocol = None
    running: bool = True

    def __post_init__(self):
        self.connected = asyncio.Event()
        self.transaction_limit = TransactionLimit(
            self.max_active_requests, self.connected
        )

    async def connect(self):
        try:
            if self.connected.is_set():
                return
        except AttributeError:
            pass
        loop = asyncio.get_running_loop()
        while self.running:
            try:
                if self.client_port:
                    sock = await self.build_reuse_socket()
                    self.transport, self.protocol = await asyncio.wait_for(
                        loop.create_connection(
                            lambda: ModbusTcpProtocol(self),
                            sock=sock,
                        ),
                        2,
                    )
                else:
                    self.transport, self.protocol = await asyncio.wait_for(
                        loop.create_connection(
                            lambda: ModbusTcpProtocol(self), self.host, self.port
                        ),
                        2,
                    )
                self.connected.set()
                return
            except (OSError, asyncio.TimeoutError) as e:
                if self.auto_reconnect_after:
                    log.warning(e)
                    log.info("Reconnecting to ModbusTCP")
                    await asyncio.sleep(self.auto_reconnect_after)
                else:
                    log.exception(e)
                    raise
            except BaseException as e:
                log.exception(e)
                raise

    async def build_reuse_socket(self):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_LINGER, struct.pack("ii", 1, 0))
        sock.settimeout(10)
        sock.bind(("", self.client_port))
        sock.setblocking(False)
        await asyncio.wait_for(
            asyncio.get_event_loop().sock_connect(sock, (self.host, self.port)),
            timeout=1,
        )
        return sock

    def _encode_packet(
        self, unit, function_code, address, trans_id, *values
    ) -> bytearray:
        packet = bytearray()
        data = encoders.from_func_code(function_code, address, *values)
        packet.extend(
            struct.pack(">HHHBB", trans_id, 0x0000, len(data) + 2, unit, function_code)
        )
        packet.extend(data)
        log.debug(("Encode: " + " ".join(f"{byt:02X}" for byt in packet)))
        return packet

    async def _request(
        self, unit: int, function_code: int, address: int, *values: int, timeout
    ):
        if unit is None:
            unit = self.default_unit_id
        if not self.running:
            raise RuntimeError("Client is stopped")
        async with self.transaction_limit:
            trans_id, fut = await self.protocol.new_transaction()
            packet = self._encode_packet(
                unit, function_code, address, trans_id, *values
            )
            self.transport.write(packet)
            timeout = timeout or self.default_timeout
            try:
                await asyncio.wait_for(fut, timeout)
            finally:
                self.protocol.transactions.pop(trans_id, None)
            return fut.result()

    async def read_coils(self, address: int, count: int, *, unit=None, timeout=None):
        req = await self._request(unit, 0x01, address, count, timeout=timeout)
        return req[:count]

    async def read_discrete_inputs(
        self, address: int, count: int, *, unit=None, timeout=None
    ):
        req = await self._request(unit, 0x02, address, count, timeout=timeout)
        return req[:count]

    async def read_holding_registers(
        self, address: int, count: int, *, unit=None, timeout=None
    ):
        return await self._request(unit, 0x03, address, count, timeout=timeout)

    async def read_input_registers(
        self, address: int, count: int, *, unit=None, timeout=None
    ):
        return await self._request(unit, 0x04, address, count, timeout=timeout)

    async def write_single_coil(
        self, address: int, value: bool, *, unit=None, timeout=None
    ):
        return await self._request(unit, 0x05, address, value, timeout=timeout)

    async def write_single_register(
        self, address: int, value: int, *, unit=None, timeout=None
    ):
        return await self._request(unit, 0x06, address, value, timeout=timeout)

    async def write_multiple_coils(
        self, address: int, *values: bool, unit=None, timeout=None
    ):
        return await self._request(unit, 0x0F, address, *values, timeout=timeout)

    async def write_multiple_registers(
        self, address: int, *values: int, unit=None, timeout=None
    ):
        return await self._request(unit, 0x10, address, *values, timeout=timeout)

    async def read_exception_status(self, unit=None, timeout=None):
        function_code = 0x07
        raise NotImplementedError

    async def diagnostics(self, sub_function, *data, unit=None, timeout=None):
        function_code = 0x08
        raise NotImplementedError

    def stop(self):
        self.running = False
        if self.transport:
            self.transport.close()

    disconnect = stop



#!/usr/bin/env python
from setuptools import setup

import sys
import os
import re

BASE_LOCATION = os.path.abspath(os.path.dirname(__file__))

REQUIRES_FILE = "requirements.txt"


def get_version():
    with open(os.path.join("aiomodbus", "__init__.py")) as f:
        return re.search("__version__ = [\"']([\w.]+)", f.read()).group(1)


def filter_comments(fd):
    no_comments = list(
        filter(lambda l: l.strip().startswith("#") is False, fd.readlines())
    )
    lines = list(filter(lambda l: l.strip().startswith("-") is False, no_comments))
    lines = list(filter(lambda l: l.strip().startswith("git+") is False, lines))
    return [l.strip() for l in lines]


def readfile(filename, func):
    try:
        with open(os.path.join(BASE_LOCATION, filename)) as f:
            data = func(f)
    except (IOError, IndexError):
        sys.stderr.write(
            """
Can't find '%s' file. This doesn't seem to be a valid release.
"""
            % filename
        )
        sys.exit(1)
    return data


def get_requires():
    return readfile(REQUIRES_FILE, filter_comments)


setup(
    name="aiomodbus",
    author="Ryan Parry-Jones",
    author_email="ryanspj+github@gmail.com",
    description="Lightweight Modbus library using asyncio",
    packages=["aiomodbus"],
    scripts=[],
    version=get_version(),
    install_requires=get_requires(),
    python_requires=">=3.7",
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Intended Audience :: Manufacturing",
        "Intended Audience :: Telecommunications Industry",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Development Status :: 2 - Pre-Alpha",
    ],
)
